/* MsgProducerAPI
 * My simple application-level API for C clients
 *
 *
 * by Fabrizio Furano, CERN, Oct 2010
 *
 */


#include <stdio.h>
#include <dlfcn.h>

#include "SEMsgProducer.hh"

extern "C" {

#include "MsgProducerCAPI.h"

}

#include <strings.h>

#include "SEMsgProtoDefs.hh"
#include "PluginLoader.hh"
#include "SimpleDebug.hh"
#include "Config.hh"










// ------------------------------------------------------------------------------------
// Plugin-related stuff
// ------------------------------------------------------------------------------------

// The plugin functionality. This function invokes the plugin loader, looking for the
// plugin where to call the hook function
SEMsgProducer *GetSEMsgProducerPluginClass_local(char *pluginPath, GetSEMsgProducerArgs) {
   PluginLoader *myLib;
   SEMsgProducer *(*ep)(GetSEMsgProducerArgs);

   // If we have no plugin path then return NULL
   if (!pluginPath || !strlen(pluginPath)) {
      Info(SimpleDebug::kMEDIUM, "GetSEMsgProducerPluginClass_local", "No plugin to load.");
      return NULL;
   }


   // Create a plugin object (we will throw this away without deletion because
   // the library must stay open but we never want to reference it again).
   Info(SimpleDebug::kMEDIUM, "GetSEMsgProducerPluginClass_local", "Loading plugin " << pluginPath);
   if (!(myLib = new PluginLoader(pluginPath))) {
      Info(SimpleDebug::kLOW, "GetSEMsgProducerPluginClass_local", "Failed loading plugin " << pluginPath);
      return NULL;
   }
   
   // Now get the entry point of the object creator
   Info(SimpleDebug::kMEDIUM, "GetSEMsgProducerPluginClass_local", "Getting entry point for plugin " << pluginPath);
   ep = (SEMsgProducer *(*)(GetSEMsgProducerArgs))(myLib->getPlugin("GetSEMsgProducer"));
   if (!ep) {
      Info(SimpleDebug::kLOW, "GetSEMsgProducerPluginClass_local", "Could not get entry point for plugin " << pluginPath);
      return NULL;
   }

   // Get the Object now
   Info(SimpleDebug::kMEDIUM, "GetSEMsgProducerPluginClass_local", "Getting class instance for plugin " << pluginPath);
   SEMsgProducer *c = ep(logobj, cfgobj, brokerURI, destname, useTopic, sessionTransacted);
   if (!c)
      Info(SimpleDebug::kLOW, "GetSEMsgProducerPluginClass_local", "Could not get class instance for plugin " << pluginPath);
   return c;

}
















// ------------------------------------
//  Initialization and finalization
// ------------------------------------
static SEMsgProducer *prod = NULL;


// Init the message producer and start it
extern "C" int MsgProd_Init(char *brokerURI, char *topicorqueue, int istopicorqueue, char *pluginname, int debuglevel) {
   if (prod) return 0;
   if (!topicorqueue) return 1;

   DebugSetLevel(debuglevel);

   // Get the entry point for the plugin that implements the product-oriented technicalities of the calls
   // e.g. the replica resolution
   // An empty string does not load any plugin, just keeps the default behavior
   prod = (SEMsgProducer *)GetSEMsgProducerPluginClass_local( pluginname,
							      SimpleDebug::Instance(),
							      Config::GetInstance(), 
							      brokerURI, topicorqueue, istopicorqueue, false);
   if (!prod) return 1;

   // Start the producer, it will try to connect.
   if (prod->Start()) {
     return 1;
   }

   return 0;
}

// Init the message producer and start it, taking the parameters from the config file
extern "C" int MsgProd_Init_cfgfile(char *cfgfile) {
  char *fname = "MsgProd_Init";

  // Process the config file
  Info(SimpleDebug::kLOW, "MsgProd_Init_cfgfile", "Starting. Config: " << cfgfile);

  if (!cfgfile || !strlen(cfgfile)) {
    Error(fname, "No config file given." << cfgfile);
    return 1;
  }

  if (CFG->ProcessFile(cfgfile)) {
    Error(fname, "Error processing config file." << cfgfile);
    return 1;
  }

  DebugSetLevel(CFG->GetLong("prodapi.debug", 1));
  long debuglevel = CFG->GetLong("prodapi.debug", 1);

  string uri = CFG->GetString("prodapi.brokerURI", "");
  string name;
  bool useTopic = false;

  name = CFG->GetString("prodapi.queue", "");

  if (name.length() == 0) {
    name = CFG->GetString("prodapi.topic", "");
    useTopic = true;
  }
    
  string pluginname = CFG->GetString("prodapi.prodplugin", "");

  return MsgProd_Init((char *)uri.c_str(), (char *)name.c_str(), useTopic, (char *)pluginname.c_str(), debuglevel);
}



// Shutdown the message producer
extern "C" void MsgProd_Terminate() {

   if (prod != NULL) {
      prod->WaitQueueEmpty(0);
      prod->Join();
      delete prod;
      prod = NULL;
   }

}


// ------------------------------------
// Utility function, to pack the voms auth stuff into a string buffer
// ------------------------------------
extern "C" void MsgProd_PackAuthstuff(char *destbuf, char *DN, char *voname, char **fqans, int nfqans) {
  destbuf[0] = '\0';
  
  if (DN)
    strcpy(destbuf, DN);

  strcat(destbuf, ",");

  if (voname)
    strcat(destbuf, voname);

  strcat(destbuf, ",");

  if (fqans && (nfqans > 0))
    for (int i = 0; i < nfqans; i++) {
      if (fqans[i])
	strcat(destbuf, fqans[i]);
      if (i < nfqans-1)
	strcat(destbuf, ",");
    }
  

  


}





// ------------------------------------
// Pre-cooked primitives to send specific messages
// ------------------------------------

// Send a notification containing a silly statement.
// Please make sure that it's silly enough.
extern "C" int MsgProd_Notifysillymessage(char *sillymessage) {
  if (!prod) {
    return -1;
  }

  // Do it 100 times, so the concept is more clear
  return prod->Say(sillymessage, 100);

}


extern "C" int MsgProd_CanNotifyNotAvailable() {
  if (!prod) {
    return 0;
  }

  return prod->CanNotifyNotAvailable();
}
extern "C" int MsgProd_NotifyNotAvailable(char *authDN, char *filename) {
  if (!prod) {
    return -1;
  }

  // Do it
  return prod->NotifyNotAvailable(authDN, filename);

}


extern "C" int MsgProd_CanNotifyAvailable() {
  if (!prod) {
    return 0;
  }

  return prod->CanNotifyAvailable();
}
extern "C" int MsgProd_NotifyAvailable(char *authDN, char *filename) {
  if (!prod) {
    return -1;
  }

  // Do it
  return prod->NotifyAvailable(authDN, filename);

}

extern "C" int MsgProd_CanNotifyChmod() {
  if (!prod) {
    return 0;
  }

  return prod->CanNotifyChmod();
}
extern "C" int MsgProd_NotifyChmod(char *authDN, char *filename, int perm) {
   if (!prod) {
    return -1;
   }

   // Do it
   return prod->NotifyChmod(authDN, filename, perm);

}

extern "C" int MsgProd_CanNotifyChown() {
  if (!prod) {
    return 0;
  }

  return prod->CanNotifyChown();
}
extern "C" int MsgProd_NotifyChown(char *authDN, char *filename, char *username, char *groupname) {
   if (!prod) {
    return -1;
   }

   // Do it
   return prod->NotifyChown(authDN, filename, username, groupname);

}

extern "C" int MsgProd_CanNotifyDeleted() {
  if (!prod) {
    return 0;
  }

  return prod->CanNotifyDeleted();
}
extern "C" int MsgProd_NotifyDeleted(char *authDN, char *filename) {
  if (!prod) {
    return -1;
  }

  // Do it
  return prod->NotifyDeleted(authDN, filename);

}


extern "C" int MsgProd_CanNotifySetACL() {
  if (!prod) {
    return 0;
  }

  return prod->CanNotifySetACL();
}
extern "C" int MsgProd_NotifySetACL(char *authDN, char *filename, char *acl) {
  if (!prod) {
    return -1;
  }

  // Do it
  return prod->NotifySetACL(authDN, filename, acl);

}

extern "C" int MsgProd_CanNotifyAddReplica() {
  if (!prod) {
    return 0;
  }

  return prod->CanNotifyAddReplica();
}
extern "C" int MsgProd_NotifyAddReplica(char *authDN,
					char      *lfn,
                                        char      *guid,
                                        int        mode,
                                        long long  size,
                                        char      *csumtype,
                                        char      *csumvalue,
                                        char      *server,
                                        char      *sfn,
                                        char      *status,
                                        char      *f_type,
                                        char      *poolname,
                                        char      *fs) {


  if (!prod) {
    return -1;
  }

  // Do it
  return prod->NotifyAddReplica(authDN, lfn, guid, mode, size, csumtype, csumvalue, server, sfn, status, f_type, poolname, fs);



}








