/* RunnableProducer
 * A simple runnable that produces messages and implements
 * async connection capabilities
 *
 *
 * by Fabrizio Furano, CERN, Jan 2011
 */

#include "RunnableProducer.hh"
#include "SimpleDebug.hh"
#include "Config.hh"
#include <typeinfo>
#include <stdio.h>
#include <string.h>

using namespace decaf::lang;

// Time to live for the producer's connection
#define PRODUCER_TTL        10

// We don't want the queue to hold too many unsent messages.
#define MESSAGEQUEUE_MAXCNT    100000

// We don't want the queue to hold messages that are too old
#define MESSAGEQUEUE_MAXTTL    86400


// A small utility class to automatically unlock Locks
class LockLocker {
private:
  decaf::util::concurrent::locks::Lock *l;
public:
  LockLocker(decaf::util::concurrent::locks::Lock *lock): l(NULL) {
    lock->lock();
    l = lock;
  }

  virtual ~LockLocker() { if (l) l->unlock(); l = NULL; }
};


int RunnableProducer::Start() {
   thr = new Thread(this);
   if (thr) thr->start();
   return (thr == NULL);
}

int RunnableProducer::Join() {
  if (!isTerminated) Term();
  thr->join();
  return 0;
}


// Wait at max millisec for the send queue to be empty
// 0 means wait forever for it to be empty
// Return true if the queue is empty, false on timeout
bool RunnableProducer::WaitQueueEmpty(int maxmillis) {
  LockLocker ll(&queueLock);

  while (sendQ.size() > 0)
    try {

      if (maxmillis == 0)
	queueEmptyCond->awaitNanos(500000000);
      else
	queueEmptyCond->awaitNanos(maxmillis*1000000);

    } catch (decaf::lang::Exception &e) {
      Info(SimpleDebug::kMEDIUM, "RunnableMsgProducer::WaitQueueEmpty", "Caught Exception" << e.getStackTraceString());
    }

  return (sendQ.size() == 0);
}



void RunnableProducer::Term() {
  // This will make the thread terminate, soon or late
   isTerminated = true;
}

void RunnableProducer::onException (const cms::CMSException &ex) {

   Info(SimpleDebug::kMEDIUM, "RunnableMsgProducer::onException", "Caught Exception" << ex.getStackTraceString());

   // Something went wrong with the connection with the broker
   // we have to trigger a reconnection attempt
   isConnected = false;

}

void RunnableProducer::run() {
  const char *fname = "RunnableProducer::run";
  time_t alivesince = 0;

  while (!isTerminated) {

    // Check if there seems to be anything ready to send
    int s = 0;
    {
      LockLocker l(&queueLock);
      s = sendQ.size();
    }

    //
    // If we are idle since a long time --> disconnect
    //
    if (isConnected && (s == 0) && ((time(0) - alivesince) > CFG->GetLong("prod.connTTL", PRODUCER_TTL))) {
      Info(SimpleDebug::kMEDIUM, fname, "Idle TTL expired. Disconnecting.");
      cleanup();
      isConnected = false;
    }


    //
    // If there's something to send and we are disconnected --> connect
    //
    if (((s > 0) || (testSendMsgs > 0)) &&
	!isConnected) {
      // Here we try to connect if there is something to do.
      Info(SimpleDebug::kMEDIUM, fname, "Trying to connect to the messaging infrastructure...");
      
      if (Connect()) 
	Error(fname, "Connection failed.")
	else {
	  Info(SimpleDebug::kMEDIUM, fname, "Connected to the messaging infrastructure.");
	  isConnected = true;
	  alivesince = time(0);
	}

      if ((isConnected) && (testSendMsgs > 0)) {
	Say((char *)"This is a test message.", testSendMsgs);
	break;
      }
      
   
    }

    //
    // Delete the messages that are too old
    // And make sure that we never keep too many messages in the queue
    // Here we don't really care if we are connected or not
    //
    {
      LockLocker l(&queueLock);
      time_t timenow = time(0);

      // Delete from the front if they are too many
      // But only if they are at least 1 minute old
      int delcnt = 0;
      while ( (int)sendQ.size() > CFG->GetLong("prod.maxqueue", MESSAGEQUEUE_MAXCNT) ) {

	if (timenow - (sendQ.front())->insTime > 60) {
	  delcnt++;
	  sendQ.erase(sendQ.begin());
	}
	else break;

      }
      if (delcnt > 0)
	Error(fname, "Queue overflow. Purged " << delcnt << " notifications. Broker troubles?");

      delcnt = 0;
      TsendQ::iterator it = sendQ.begin();
      timenow = time(0);
      while(it != sendQ.end()) {
	if ( timenow - (*it)->insTime > CFG->GetLong("prod.queueTTL", MESSAGEQUEUE_MAXTTL) )
	  it = sendQ.erase(it);
	else
	  it++;
      }
      if (delcnt > 0)
	Error(fname, "Purged " << delcnt << " older notifications. Broker troubles?");

    }

    //
    // If we are effectively connected --> send the messages
    //
    if (isConnected) {

      
      if (!isTerminated && (s > 0)) {
	Info(SimpleDebug::kHIGH, fname, "Still " << s << " notifications to send.");
	localMsg *lm = 0;
	localMsg *llm = 0;
	int mergecnt = 0;
	string mergedcmd, mergedauthDN, authDN;


	// Lock the queue and take the first message to process
	{
	  LockLocker l(&queueLock);
	  if (sendQ.size() <= 0) {
	    s = 0;
	    //break;
	  }
	  
	  lm = sendQ[0];

	  if (!lm) {
	    Error(fname, "Found null msg in the queue?!?");
	    sendQ.erase(sendQ.begin());
	    queueEmptyCond->signal();
	    //break;
	  }

	mergedcmd = lm->hdrstr[PROTO_CMD_NAME];
	mergedauthDN = "";
	if (lm->hdrstr.find(PROTO_AUTHDN) != lm->hdrstr.end())
	  mergedauthDN = lm->hdrstr[PROTO_AUTHDN];

	// Now scan the queue forward, to see if we can find similar messages
	// to gather into a bulk one
	TsendQ::iterator it = sendQ.begin();
	llm = new localMsg((char *)mergedcmd.c_str());
	
	while(it != sendQ.end()) {
	  char s1[1024], s2[1024];
	  sprintf(s2, "%.4d", mergecnt+1);
	 
	  authDN = "";
	  if ((*it)->hdrstr.find(PROTO_AUTHDN) != (*it)->hdrstr.end())
	    authDN = (*it)->hdrstr[PROTO_AUTHDN];

	  if ( ((*it)->hdrstr[PROTO_CMD_NAME] == mergedcmd) && (authDN == mergedauthDN) ) {
	    // We found an element to merge
	    // Just add the elements of its body concatenating the keys with cnt

	    
	    for (TStrStrMap::iterator it1 = (*it)->bodystr.begin(); it1 != (*it)->bodystr.end(); it1++) {
	      strcpy(s1, it1->first.c_str());
	      strcat(s1, s2);
	      llm->bodystr[s1] = it1->second;
	    }
	    for (TStrLongMap::iterator it1 = (*it)->bodylong.begin(); it1 != (*it)->bodylong.end(); it1++) {
	      strcpy(s1, it1->first.c_str());
	      strcat(s1, s2);
	      llm->bodylong[s1] = it1->second;
	    }
	    for (TStrLonglongMap::iterator it1 = (*it)->bodylonglong.begin(); it1 != (*it)->bodylonglong.end(); it1++) {
	      strcpy(s1, it1->first.c_str());
	      strcat(s1, s2);
	      llm->bodylonglong[s1] = it1->second;
	    }

	    // Mark the message as being sent
	    // So that it survives in the queue if there are errors
	    (*it)->beingSent = true;
	    it++;

	    mergecnt++;
	    lm = NULL;
	    if (mergecnt >= 1000) break;
	  }
	  else it++;


	}


	
	} // Close lock

	cms::Message *m;

	llm->hdrlong.insert(TStrLongPair(PROTO_NENTRIES, mergecnt));
	if (mergedauthDN != "")
	  llm->hdrstr.insert(TStrStrPair(PROTO_AUTHDN, mergedauthDN));


	//
	// The (bulk) message has been composed, xlate it into the CMSMessage we have to send
	//
	m = GetMapMessage(llm);
	SignMessage((cms::MapMessage*)m);

	//CheckMessage((cms::MapMessage*)m);

	delete llm;

	if (m) {
	  // Yeah send this message. A failure here throws the message away
	  if (producer) {
	    try {
	      Info(SimpleDebug::kHIGH, fname, "Sending message...");
	      producer->send( m );
	      Info(SimpleDebug::kLOW, fname, "Message sent.");

	      // Now delete the pending messages from the queue
	      // Here we should eventually ACK the producer plugin
	      int d = 0;
	      TsendQ::iterator it = sendQ.begin();
	      while(it != sendQ.end()) {
		if ((*it)->beingSent) {
		  delete (*it);
		  sendQ.erase(it);
		  d++;
		}
		else
		  it++;
	      }
	      Info(SimpleDebug::kHIGH, fname, "Successfully processed " << d << " sub-messages");
	      
	    } catch (cms::CMSException &e) {
	      Error(fname, "Exception while sending message:" << e.getMessage() << " - " << e.getStackTraceString());
	      isConnected = false;
	    }
	   
	  }
	  else {
	    Error(fname, "Found null message producer?!?");
	    isConnected = false;
	    //break;
	  }

	  delete m;

	}
	

	// Notify that the queue might be empty
	{
	  LockLocker l(&queueLock);
	  s = sendQ.size();
	}
	queueEmptyCond->signal();
	
	
      }

    }

    // Sloppy but good enough implementation of an interruptible sleep
    for (int i = 1; !isTerminated && (i < 5); i++)
      thr->sleep(100);

  }
  Info(SimpleDebug::kMEDIUM, fname, "Exiting.");
};





// From an instance of myMsg, create a message to be sent to the broker
cms::Message *RunnableProducer::GetMapMessage(localMsg *msg) {
   const char *fname = "RunnableProducer::GetMapMessage";

   // A new message
   MapMessage* m = NULL;

   if (session)
     m = session->createMapMessage();

   if (!m) {
      Error(fname, "Cannot create MapMessage.");
      return NULL;
   }

   // Set the string fields of the header
   for (TStrStrMap::iterator i = msg->hdrstr.begin(); i != msg->hdrstr.end(); i++) {
     Info(SimpleDebug::kHIGHEST, fname,
	  "Setting header str " << i->first << "=" << i->second);

     m->setStringProperty(i->first, i->second);
   }
   // Set the long fields of the header
   for (TStrLongMap::iterator i = msg->hdrlong.begin(); i != msg->hdrlong.end(); i++) {
     Info(SimpleDebug::kHIGHEST, fname,
	  "Setting header long " << i->first << "=" << i->second);

     m->setIntProperty(i->first, i->second);
   }
   // Set the longlong fields of the header
   for (TStrLonglongMap::iterator i = msg->hdrlonglong.begin(); i != msg->hdrlonglong.end(); i++) {
     Info(SimpleDebug::kHIGHEST, fname,
	  "Setting header longlong " << i->first << "=" << i->second);

     m->setLongProperty(i->first, i->second);
   }

   // Set the string fields of the body
   for (TStrStrMap::iterator i = msg->bodystr.begin(); i != msg->bodystr.end(); i++) {
     Info(SimpleDebug::kHIGHEST, fname,
	  "Setting body str " << i->first << "=" << i->second);

     m->setString(i->first, i->second);
   }
   // Set the long fields of the body
   for (TStrLongMap::iterator i = msg->bodylong.begin(); i != msg->bodylong.end(); i++) {
     Info(SimpleDebug::kHIGHEST, fname,
	  "Setting body long " << i->first << "=" << i->second);

     m->setInt(i->first, i->second);
   }
   // Set the longlong fields of the body
   for (TStrLonglongMap::iterator i = msg->bodylonglong.begin(); i != msg->bodylonglong.end(); i++) {
     Info(SimpleDebug::kHIGHEST, fname,
	  "Setting body longlong " << i->first << "=" << i->second);

     m->setLong(i->first, i->second);
   }


   return m;
}

