/* SEMsgSecurity
 * A class that encapsulates the security stuff of SEMsg
 *
 *
 *
 * by Fabrizio Furano, CERN, Jan 2011
 */

#include <decaf/lang/System.h>

#include "SEMsgSecurity.hh"
#include "SimpleDebug.hh"
#include "Config.hh"
#include "SEMsgProtoDefs.hh"
#include "SEMsgConfig.h"

#include <stdio.h>
#include <stdlib.h>
#include <openssl/evp.h>
#include <openssl/pem.h>
#include <openssl/rsa.h>
#include <openssl/err.h>
#include <openssl/x509v3.h>
#include <openssl/ssl.h>

#include <vector>

// Encode an array of bytes to base64
char *Tobase64(const unsigned char *input, int length)
{
  BIO *bmem, *b64;
  BUF_MEM *bptr;

  b64 = BIO_new(BIO_f_base64());
  bmem = BIO_new(BIO_s_mem());
  b64 = BIO_push(b64, bmem);
  BIO_write(b64, input, length);

  if (BIO_flush(b64) <= 0) {
    BIO_free_all(b64);
    return NULL;
  }

  BIO_get_mem_ptr(b64, &bptr);

  char *buff = (char *)malloc(bptr->length);
  memcpy(buff, bptr->data, bptr->length-1);
  buff[bptr->length-1] = 0;

  BIO_free_all(b64);

  return buff;
}





// Text representation of a public key
char *Pubkey2Str(EVP_PKEY *input)
{
  BIO *bmem, *b;
  BUF_MEM *bptr;
  int length;

  b = BIO_new(BIO_f_buffer());
  bmem = BIO_new(BIO_s_mem());
  b = BIO_push(b, bmem);

  length = PEM_write_bio_PUBKEY(b, input);


  if (BIO_flush(b) <= 0) {
    BIO_free_all(b);
    return NULL;
  }

  BIO_get_mem_ptr(b, &bptr);

  char *buff = (char *)malloc(bptr->length);
  memcpy(buff, bptr->data, bptr->length-1);
  buff[bptr->length-1] = 0;

  BIO_free_all(b);

  return buff;
}




// Get a pubkey from a string representation
EVP_PKEY *Str2Pubkey(char *input)
{
  BIO *bmem, *b;
  EVP_PKEY *k;

  b = BIO_new(BIO_f_buffer());
  bmem = BIO_new(BIO_s_mem());
  b = BIO_push(b, bmem);

  BIO_puts(bmem, input);
  k = PEM_read_bio_PUBKEY(b, NULL, NULL, NULL);

  BIO_free_all(b);

  return k;
}


// Get a cert from a string
X509 *Str2Cert(char *input) {
  BIO *bmem, *b;
  X509 *k;

  b = BIO_new(BIO_f_buffer());
  bmem = BIO_new(BIO_s_mem());
  b = BIO_push(b, bmem);

  BIO_puts(bmem, input);
  k = PEM_read_bio_X509(b, NULL, NULL, NULL);

  BIO_free_all(b);

  return k;
}


// Verify the certificate
int SEMsgSecurity::check_cert1(X509 *cert) {
  const char *fname = "SEMsgSecurity::check_cert1";
  int res = 0;

  X509_STORE_CTX *storeCtx = X509_STORE_CTX_new();
  if (!storeCtx) {
    Error(fname, "Unable to create store ctx.");
    return -1;
  }

  X509_STORE_set_flags(m_store, 0);
  X509_STORE_CTX_init(storeCtx, m_store, cert, NULL);

  if (X509_verify_cert(storeCtx) == 1) {
    Info(SimpleDebug::kHIGH, fname, "Verified");
  }
  else {

    Info(SimpleDebug::kHIGH, fname, "Verification error: " <<
	 X509_verify_cert_error_string(X509_STORE_CTX_get_error(storeCtx)));

    while(int e = ERR_get_error())
      Info(SimpleDebug::kHIGH, fname, "Verification error: " <<
	   ERR_error_string(e, NULL) );
    res = 1;
  }


  X509_STORE_CTX_free(storeCtx);
  return res;
}




// Check the cert given by the peer, verify that
// this is his own certificate. Skip this check only if host is null or empty
int SEMsgSecurity::check_cert2(X509 *cert, char *name) {
  X509_NAME *subj;
  char data[256];
  int extcount;
  int ok = 0;

  if ((extcount = X509_get_ext_count(cert)) > 0) {
    int i;
    
    for (i = 0; i < extcount; i++) {
      const char *extstr;
      X509_EXTENSION *ext;

      ext = X509_get_ext(cert, i);
      extstr = OBJ_nid2sn(OBJ_obj2nid(X509_EXTENSION_get_object(ext)));

      if (!strcmp(extstr, "subjectAltName")) {
	int j;
	const unsigned char *data;
	STACK_OF(CONF_VALUE) *val;
	CONF_VALUE *nval;
	X509V3_EXT_METHOD *meth;

	if (!(meth = (X509V3_EXT_METHOD *)X509V3_EXT_get(ext))) break;

	data=ext->value->data;

	val=meth->i2v(meth,
		      meth->d2i(NULL, &data, ext->value->length),
		      NULL);

	for (j = 0; j < sk_CONF_VALUE_num(val); j++) {
	  nval = sk_CONF_VALUE_value(val, j);
	  if (!strcmp(nval->name, "DNS") && !strcmp(nval->value, name)) {
	    ok = 1;
	    break;
	  }
	}
      }

      if (ok) break;
    }
  }

  if (!ok && (subj = X509_get_subject_name(cert)) &&
      X509_NAME_get_text_by_NID(subj, NID_commonName, data, 256) > 0) {
    data[255] = 0;
    if(strcasecmp(data, name) != 0)
      goto err_occurred;
  }

  return X509_V_OK;

 err_occurred:

  if (cert)
    X509_free(cert);

  return X509_V_ERR_APPLICATION_VERIFICATION;

}




// Given a message, build up the buffer needed to univocally validate it
// against the signature
size_t SEMsgSecurity::BuildVfyBuf(const cms::MapMessage* m) {
  char *fname = "SEMsgSecurity::BuildVfyBuf";

  // Build a vector of items, containing the field names from the message body
  std::vector<std::string> v = m->getMapNames();
  
  // Also consider the command name and NENTRIES (got from the header)
  v.push_back(PROTO_CMD_NAME);
  v.push_back(PROTO_NENTRIES);
  v.push_back(PROTO_SENDERDN);
  v.push_back(PROTO_AUTHDN);

  // Sort the vector
  sort(v.begin(), v.end());

  // Concatenate all the strings into the buffer
  std::vector <std::string>::iterator it = v.begin();
  vfybuf[0] = 0;
  int totlen = 0;
  while (it != v.end()) {
    int l = it->length();
    // Double the buffer if it is not sufficient
    while (totlen + l >= vfybuflen-1) {
      
      char *b = (char *)realloc(vfybuf, vfybuflen*2);
      if (b) {
	vfybuf = b;
	vfybuflen *= 2;
      }
      else {
	Error(fname, "Out of memory allocating verify buffer.");
	vfybuf[0] = 0;
	return 0;
      }
    }

    // Append the name of the field
    strcat(vfybuf, it->c_str());
    std::string s;
    try {
      s = m->getString(*it);
    } catch (cms::CMSException& e) {

      if ((*it == PROTO_CMD_NAME) || (*it == PROTO_NENTRIES) || (*it == PROTO_SENDERDN) || (*it == PROTO_AUTHDN)) {

	try {
	  s = m->getStringProperty(*it);
	} catch (cms::CMSException& e) {
	  Error(fname, "Field " << *it << " not found.");
	  return 0;
	}


      }

    }
    
    // Append the value of the fields

    strcat(vfybuf, s.c_str());
    totlen += l;
    totlen += s.length();
	   
    it++;
  }





  return totlen;
  
}









SEMsgSecurity::SEMsgSecurity() {
  rsa_pkey = NULL;
  pkey = EVP_PKEY_new();
  pubkey = NULL;
  //sig = (unsigned char *)malloc(EVP_PKEY_size(pkey));
  //Info(SimpleDebug::kMEDIUM, "SEMsgSecurity::SEMsgSecurity", EVP_PKEY_size(pkey));
  sig = (unsigned char *)malloc(256);
  siglen = 0;

  m_store = NULL;
  m_lookup = NULL;

  vfybuf = (char *)malloc(1024*1024);
  vfybuflen = 1024*1024;

  // Build the fast lookup for the whitelist, from the config
  char buf[1024];
  int i = 0;
  do {
    CFG->ArrayGetString("whitelist.allowsender", buf, i);
    if (buf[0])
      senderWL[buf].push_back("/");
    i++;
  } while (buf[0]);

  init_ok = false;
}

SEMsgSecurity::~SEMsgSecurity() {

  if(m_store != NULL) {
    X509_STORE_free(m_store);
    m_store = NULL;
  }


  if (pkey)
    EVP_PKEY_free(pkey);

  if (pubkey)
    EVP_PKEY_free (pubkey);

  if (sig)
    free(sig);

  if (init_ok)
    EVP_MD_CTX_cleanup(&ctx);
    EVP_cleanup();

  ERR_free_strings();

}

int SEMsgSecurity::Init(char *cfgprefix) {

  char *fname = "SEMsgSecurity::Init";

  std::string pfx = cfgprefix;
  std::string s;

  // Get the grid cert+key in its preferred PEM format.
  std::string certfile;
  std::string certkeyfile;
  s = pfx + ".cert";
  if (getenv("X509_USER_CERT"))
    certfile = CFG->GetString((char *)s.c_str(), getenv("X509_USER_CERT"));
  else
    certfile = CFG->GetString((char *)s.c_str(), "");

  s = pfx + ".certkey";
  if (getenv("X509_USER_KEY"))
    certkeyfile = CFG->GetString((char *)s.c_str(), getenv("X509_USER_KEY"));
  else
    certkeyfile = CFG->GetString((char *)s.c_str(), "");



  // Set the ActiveMQ SSL connection parameters
  s = pfx + ".trustStore";
  decaf::lang::System::setProperty( "decaf.net.ssl.trustStore",
				    CFG->GetString((char *)s.c_str(), "" ));
  s = pfx + ".trustStorePassword";
  decaf::lang::System::setProperty( "decaf.net.ssl.trustStorePassword",
				    CFG->GetString((char *)s.c_str(), "" ));
  s = pfx + ".keyStore";
  decaf::lang::System::setProperty( "decaf.net.ssl.keyStore",
				    CFG->GetString((char *)s.c_str(), "" ));
  s = pfx + ".keyStorePassword";
  decaf::lang::System::setProperty( "decaf.net.ssl.keyStorePassword", 
				    CFG->GetString((char *)s.c_str(), "" ));

  // Init the engine with our private key
  FILE *myf = fopen(certkeyfile.c_str(), "r");
  if (!myf) {
    Error(fname, "Error opening RSA Private Key File (" << errno << "): " << strerror(errno));
    return 2;
  }

  SSL_load_error_strings();
  OpenSSL_add_all_ciphers();
  OpenSSL_add_all_algorithms();

  if (!PEM_read_RSAPrivateKey(myf , &rsa_pkey, NULL, NULL)) {
    Error(fname, "Error loading RSA Private Key File." << ERR_error_string(ERR_get_error(), NULL));
    fclose(myf);
    return 2;
  }

  fclose(myf);

  if (!EVP_PKEY_assign_RSA(pkey, rsa_pkey)) {
    Error(fname, "EVP_PKEY_assign_RSA: failed on the private key.");
    return 3;
  }


  // Init the engine with our pubkey, got from the cert
  myf = fopen(certfile.c_str(), "r");
  if (!myf) {
    Error(fname, "Error opening certificate (" << errno << "): " << strerror(errno));
    return 2;
  }

  X509 *x509=PEM_read_X509(myf, NULL, NULL, NULL);
  if (!x509) {
    Error(fname, "Error loading certificate." << ERR_get_error());
    fclose(myf);
    return 2;
  }


  // Keep a string version of the cert, to be able to quickly send it
  rewind(myf);
  while (!feof(myf)) {
    char buf[1025];
    size_t l = fread((void *)buf, 1, sizeof(buf)-1, myf);
    buf[l] = 0;
    mycert += buf;
  }

  fclose(myf);


  // Get my DN
  X509_NAME *nm = X509_get_subject_name(x509);
  char namebuf[256];
  memset(namebuf, 0, sizeof(namebuf));
  if (nm) {
    X509_NAME_oneline(nm, namebuf, sizeof(namebuf));
    myDN = namebuf;
  }

  if (myDN.length() == 0) {
    myDN += getenv("USER");
    myDN += "@";
    myDN += getenv("HOST");
  }


  Info(SimpleDebug::kMEDIUM, fname, "My DN: '" << myDN <<"'");


  pubkey=X509_get_pubkey(x509);
  if (!pubkey) {
    Error(fname, "X509_get_pubkey: failed on the public key from the cert.");
    return 3;
  }

  char *o = Pubkey2Str(pubkey);
  strpubkey = o;
  free(o);
  Info(SimpleDebug::kMEDIUM, fname, "Public key: " << strpubkey);


  // Init the certstore used to verify certificates
  m_store = X509_STORE_new();
  X509_STORE_set_default_paths(m_store);
  
  m_lookup=X509_STORE_add_lookup(m_store,X509_LOOKUP_hash_dir());
  
  s = pfx + ".CAdir";
  if (!X509_STORE_load_locations(m_store,
				 NULL,
				 CFG->GetString((char *)s.c_str(), "/etc/grid-security/certificates").c_str()) ) {
    
    Error(fname, "Error in X509_STORE_load_locations");
  }

  EVP_MD_CTX_init(&ctx);
  init_ok = true;
  return 0;

}

// Computes the signature for a message (avoiding the signature field)
int SEMsgSecurity::CalcSignature(  const cms::MapMessage* m, std::string &sigstr) {
  char *fname = "SEMsgSecurity::CalcSignature";
  if (!init_ok) return 0;

  // Build the data to verify...
  size_t len = BuildVfyBuf(m);
  Info(SimpleDebug::kHIGH, fname, "Going to sign " << len <<" bytes.");

  if (!EVP_SignInit_ex(&ctx, EVP_sha1(), NULL)) {
      Error(fname, "EVP_SignInit: failed.");
      return 3;
    }
  
    // Process len bytes from buffer
    if (!EVP_SignUpdate(&ctx, vfybuf, len)) {
      Error(fname, "EVP_SignUpdate: failed.");
      return 3;
    }
  
    // Put the signature into sig, get its final length
    if (!EVP_SignFinal(&ctx, sig, &siglen, pkey)) {
      Error(fname, "EVP_SignUpdate: failed.");
      return 3;
    }
  
    sigstr.clear();
    for (unsigned int i = 0; i < siglen; i++) {
      char buf[256];
      sprintf(buf, "%02x", sig[i]);
      sigstr+=(buf);
      
    }


    Info(SimpleDebug::kMEDIUM,
	 fname,
	 "Signature of " << len <<
	 " bytes (" << sigstr.length() <<
	 ")" << sigstr);

    return 0;
}

// Signs the message
int SEMsgSecurity::SignMessage( cms::MapMessage* m ) {
  if (!init_ok) return 0;

  std::string s;


  // Put my pub certificate into the message. This will be refined so that the cert is not
  // sent in each message, but only from time to time
  m->setString(PROTO_X509CERT, mycert);

  // Add a field with the sender's DN
  m->setStringProperty(PROTO_SENDERDN, myDN);

  CalcSignature( m, s );

  // Put the signature into the message
  m->setStringProperty(PROTO_SIGNATURE, s);

  return 0;

}

// Verifies the signature in the given message
int SEMsgSecurity::CheckSignature(  const cms::MapMessage* m ) {
  char *fname = "SEMsgSecurity::CheckSignature";
  if (!init_ok) return 0;


  // Get the signature from the message, convert into binary buffer
  char msgsig_buf[4096];
  int msgsig_len = 0;
  std::string sig;
  EVP_PKEY *pubk = NULL;

  try {
    sig = m->getStringProperty(PROTO_SIGNATURE);
  } catch (cms::CMSException& e) {

    Error(fname, "Signature not available. Message rejected.");
    return 1;

  }

  unsigned int v = 0;
  for (unsigned int i = 0; i < sig.length()/2; i++) {
    sscanf(&sig.c_str()[i*2], "%02x", &v);
    msgsig_buf[i] = v;
  }
  msgsig_len = sig.length()/2;


  // Get the certificate from the message, process it
  std::string strcert;


  try {
    strcert= m->getString(PROTO_X509CERT);
  } catch (cms::CMSException& e) {}

  if (strcert.length() > 0) {
    Info(SimpleDebug::kHIGH, fname, "Got certificate:" << strcert);
    X509 *x509=Str2Cert((char *)strcert.c_str());
    if (!x509) {
      Error(fname, "Error loading certificate." << ERR_get_error());
      return 2;
    }

    // Verify the certificate chain
    if (check_cert1(x509)) {
      Error(fname, "Error verifying certificate chain:" << strcert);
      return 3;
    }

    // Verify the certificate DN
    X509_NAME *nm = X509_get_subject_name(x509);
    char namebuf[256];
    memset(namebuf, 0, sizeof(namebuf));
    if (nm) 
      X509_NAME_oneline(nm, namebuf, sizeof(namebuf));
    
    std::string sender;
    try {
      sender = m->getStringProperty(PROTO_SENDERDN);
    } catch (cms::CMSException& e) {
      Error(fname, "Message has no sender information. Rejected. cert:" << strcert);
      return 3;
    }

    if (sender != namebuf) {
      Error(fname, "The message sender '" << sender << "' does not match its certificate's DN: " << namebuf);
      return 3;
    }
    

    // Get the pubkey from the certificate
    pubk = X509_get_pubkey(x509);

    if (!pubk) {
      Info(SimpleDebug::kHIGH, fname, "Failed to get pubkey from peer cert.");
      
      while(int e = ERR_get_error())
	Info(SimpleDebug::kHIGH, fname, "Error: " <<
	     ERR_error_string(e, NULL) );
    }



  }
  else {
    Error(fname, "No certificate found.");
      return 3;
    }





  // Build the data to verify...
  size_t len = BuildVfyBuf(m);






  // Verify the signature against the received pubkey
  
  if (!EVP_VerifyInit_ex(&ctx, EVP_sha1(), NULL))  {
    Error(fname, "EVP_VerifyInit: failed.");
    return 3;
  }

  if (!EVP_VerifyUpdate (&ctx, vfybuf, len))  {
    Error(fname, "EVP_VerifyUpdate: failed.");
    return 3;
  }

  int res = EVP_VerifyFinal (&ctx,
			     (const unsigned char*)msgsig_buf,
			     msgsig_len,
			     pubk);
  

  EVP_PKEY_free(pubk);

  if (res != 1)  {
    if (res == 0) {
      Error(fname, "This message has an invalid signature.");
      return 4;
    }
    if (res == -1) {
      Error(fname, "Internal error in EVP_VerifyFinal");
      return 5;
    }
    Error(fname, "EVP_VerifyFinal: unknown return value.");
    return 6;
  }
  
  Info(SimpleDebug::kHIGH, fname, "Message verification OK!");
  // Success!
  return(0);
}


// Validates the signature of a message
int SEMsgSecurity::CheckMessage( const cms::MapMessage* m ) {
  if (!init_ok) return 0;
  char *fname = "SEMsgSecurity::CheckMessage";


  std::string sender, authDN;

  try {
    sender = m->getStringProperty(PROTO_SENDERDN);
  } catch (cms::CMSException& e) {
    Error(fname, "Message has no sender information. Rejected.");
    return 3;
  }

  try {
    authDN = m->getStringProperty(PROTO_AUTHDN);
  } catch (cms::CMSException& e) {
    Error(fname, "Message has no authDN information. Rejected.");
    return 3;
  }

  if (authDN.size() < 1) {
    Error(fname, "Message has an empty authDN information. Rejected.");
    return 3;
  }

  // Verify that the sender is in the whitelist, if any
  if (senderWL.size() > 0) {
    // The whitelist is not empty

    if (senderWL.find(sender) == senderWL.end()) {
      Error(fname, "Unauthorized sender. Message rejected.");
      return 3;
    }

  }

  return CheckSignature(m);

}
