/* SEMsgdaemon
 * A daemon starting a message consumer
 * This is a small app that:
 * - loads the proper plugin, whose name is given in the command line
 *   - if no plugin is specified then the default one will be used
 * - instantiates that consumer class, that will
 *   - wait for messages
 *   - invoke the corresponding actions on the plugin
 * - Can take all the config parameters from a config file, parsed
 *   by applying envvar substitutions
 *
 *
 * by Fabrizio Furano, CERN, Jan 2011
 *
 *
 * Parameters:
 *    TestSEMsgConsumer <URI> [-c configfile] [-b] [--allownosec] [-t <topicname>] [-q <queuename>] [-p <PluginName>] [-d <DebugLevel>]
 *
 *    An example URI is: "tcp://gridmsg001.cern.ch:6166?wireFormat=openwire&transport.commandTracingEnabled=true"
 *
 *
 *
 *
 */


#include "SEMsgConsumer.hh"
#include "Config.hh"
#include "SEMsgConfig.h"
#include <stdlib.h>
#include <string.h>
#include <vector>

using namespace std;

int main (int argc, char *argv[]) {

   if (argc < 2) {
      cout << endl <<
         "Usage: " << endl <<
         "SEMsgdaemon <URI> [-c configfile] [-b] [--allownosec] [-t <topicname>] [-q <queuename>] [-p <PluginName>] [-d <DebugLevel>]" << endl;
      return 1;
   }

   string name = "";
   string pluginName = "";
   bool useTopic = false;
   int dbg = -100;
   bool detach = false;
   string cfgfile = "";
   bool allownosec = false;

   for (int i = 2; i < argc; i++) {
      if (!strcmp(argv[i], "-t")) {
         name = argv[++i];
         useTopic = true;
         continue;
      }
      if (!strcmp(argv[i], "-q")) {
         name = argv[++i];
         useTopic = false;
         continue;
      }
      if (!strcmp(argv[i], "-p")) {
         pluginName = argv[++i];
         continue;
      }
      if (!strcmp(argv[i], "-d")) {
         dbg = atoi(argv[++i]);
	 DebugSetLevel(dbg);
         continue;
      }
      if (!strcmp(argv[i], "-c")) {
 	 cfgfile = argv[++i];
         continue;
      }
      if (!strcmp(argv[i], "-b")) {
 	 detach = true;
         continue;
      }
      if (!strcmp(argv[i], "--allownosec")) {
 	 allownosec = true;
         continue;
      }
   }


   // Background mode if desired
   if (detach) daemon(1, 0);


   Info(SimpleDebug::kLOW, argv[0], "Starting. Version: " <<
	SEMsg_VERSION_MAJOR << "." <<
        SEMsg_VERSION_MINOR << "." <<
        SEMsg_VERSION_PATCH);
   


   // Process the config file
   if (cfgfile.length() > 0) {
     if (CFG->ProcessFile((char *)cfgfile.c_str())) {
       Error(argv[0], "Error processing config file " << cfgfile);
       return 1;
     }
   }

   // Now fixup the values. The command line has precedence over the config file
   if (dbg == -100) {
     DebugSetLevel(CFG->GetLong("consd.debug", 1));
		   }

   if (pluginName.length() == 0)
     pluginName = CFG->GetString("consd.consplugin", "");

   if (name.length() == 0) {
     name = CFG->GetString("consd.queue", "");
     useTopic = false;
   }

   if (name.length() == 0) {
     name = CFG->GetString("consd.topic", "");
     useTopic = true;
   }

   string uri = CFG->GetString("consd.brokerURI", "");
   if (strcmp(argv[1], "-"))
       uri = argv[1];
   Info(SimpleDebug::kLOW, argv[0], "URI: " << uri);

   if (useTopic) {
     Info(SimpleDebug::kLOW, argv[0], "Topic: " << name);
   }
   else {
     Info(SimpleDebug::kLOW, argv[0], "Queue: " << name);
   }

   Info(SimpleDebug::kLOW, argv[0], "Plugin: " << pluginName);

   activemq::library::ActiveMQCPP::initializeLibrary();

   // Get the plugin entry point class
   SEMsgConsumer *c = (SEMsgConsumer *)GetSEMsgConsumerPluginClass( (char *)pluginName.c_str(),
                                                            uri, name, useTopic, false);
   if (!c) {
     Error(argv[0], "Error getting plugin entry point.");
     return 1;
   }

   // Exit if the security is OFF, unless the force switch was specified
   if (!c->GetSecurityStatus() && !allownosec) {
     Error(argv[0], "Running this server with no security is not currently allowed.");
     return 1;
   }

   // We have nothing else to do, hence we run this instead of putting it into a thread
   c->run();

   activemq::library::ActiveMQCPP::shutdownLibrary();

   return 0;
}
