

/* SimpleConsumer
 * A simple, encapsulated message consumer
 *
 *
 * by Fabrizio Furano, CERN, Oct 2010
 *
 * Originally taken from HelloWorldConsumer in the ActiveMQcpp
 * examples. The original license follows:
 *
 *
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "SimpleConsumer.hh"
#include "SimpleDebug.hh"
#include <activemq/library/ActiveMQCPP.h>
#include <activemq/core/ActiveMQConnectionFactory.h>

#include <string>
#include <iostream>
#include <memory>

using namespace std;
using namespace cms;

// It does what it says
int SimpleConsumer::Connect() {
  cleanup();

  try
    {

      // Create a ConnectionFactory
      auto_ptr<ConnectionFactory> connectionFactory(
						    ConnectionFactory::createCMSConnectionFactory( brokerURI ) );
      
      // Create a Connection
      connection = connectionFactory->createConnection();
      connection->start();
      connection->setExceptionListener(this);
      
      // Create a Session
      if( this->sessionTransacted == true ) {
	session = connection->createSession( Session::SESSION_TRANSACTED );
      } else {
	session = connection->createSession( Session::AUTO_ACKNOWLEDGE );
      }
      
      // Create the destination (Topic or Queue)
      if( useTopic ) {
	destination = session->createTopic( destname );
      } else {
	destination = session->createQueue( destname );
      }
      
      // Create a MessageConsumer from the Session to the Topic or Queue
      consumer = session->createConsumer( destination );
      
      consumer->setMessageListener( this );
      
      return 0;

    } catch (CMSException& e) {
       cleanup();
       Error("SimpleConsumer::Connect",
	     "CMSException catched while connecting to " << brokerURI << " Exception:" << e.getStackTraceString());

       return 1;
  }

}



// Called from the consumer since this class is a registered MessageListener.
void SimpleConsumer::onMessage( const Message* message ) throw() {

   static int count = 0;

   try
   {
      count++;
      const TextMessage* textMessage =
         dynamic_cast< const TextMessage* >( message );
      string text = "";

      if( textMessage != NULL ) {
         text = textMessage->getText();
      } else {
         text = "NOT A TEXTMESSAGE!";
      }

      //cout << "Message #" << count << " Received: '" << text << "'" << endl;

   } catch (CMSException& e) {
     Error("SimpleConsumer::onMessage",
	     "CMSException catched while processing msg:" << e.getStackTraceString());
   }

   // Commit all messages.
   CommitMessages();


}


int SimpleConsumer::CommitMessages() {
  // Commit all messages.
  if( this->sessionTransacted ) {
    try {
      Info(SimpleDebug::kHIGH, "SimpleConsumer::CommitMessages",
	    "Committing...");
      session->commit();
    } catch (cms::CMSException& e) {
      Error("SimpleConsumer::CommitMessages",
	    "CMSException catched. Problem committing. Exception:" << e.getStackTraceString());
      return 1;
    }
  }

  return 0;
}





// If something bad happens you see it here as this class is also been
// registered as an ExceptionListener with the connection.
void SimpleConsumer::onException( const CMSException& ex AMQCPP_UNUSED) throw() {
  cout << "CMS Exception occurred:" << ex.getMessage() << " - " << ex.getStackTraceString() << endl;
  
}

void SimpleConsumer::cleanup(){
  char *fname= "SimpleConsumer::cleanup";

   //*************************************************
   // Always close destination, consumers and producers before
   // you destroy their sessions and connection.
   //*************************************************

   // Destroy resources.
   try{
      if( destination != NULL ) delete destination;
   }catch (CMSException& e) {
     Error(fname,
	   "CMSException catched while deleting destination:" << e.getStackTraceString());
   }
   destination = NULL;

   try{
     if( consumer != NULL ) delete consumer;
   }catch (CMSException& e) {
     Error(fname,
	   "CMSException catched while deleting consumer:" << e.getStackTraceString());
   }
   consumer = NULL;

   // Close open resources.
   try{
      if( session != NULL ) session->close();
      if( connection != NULL ) connection->close();
   }catch (CMSException& e)  {
     Error(fname,
	   "CMSException catched while closing session/connection" << e.getStackTraceString());
   }

   // Now Destroy them
   try{
      if( session != NULL ) delete session;
   }catch (CMSException& e)   {
     Error(fname,
	   "CMSException catched while deleting session" << e.getStackTraceString());
   }

   session = NULL;

   try{
      if( connection != NULL ) delete connection;
   }catch (CMSException& e) {
     Error(fname,
	   "CMSException catched while deleting connection" << e.getStackTraceString());
   }
   connection = NULL;
}

