/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include "SEMsgConfig.h"
#include "SimpleProducer.hh"
#include "SimpleDebug.hh"


#include <decaf/util/concurrent/CountDownLatch.h>
#include <decaf/lang/Integer.h>
#include <decaf/util/Date.h>

#include <cms/TextMessage.h>
#include <cms/BytesMessage.h>
#include <cms/MapMessage.h>
#include <cms/ExceptionListener.h>

#include <memory>

using namespace activemq::core;
using namespace decaf::util::concurrent;
using namespace decaf::util;
using namespace decaf::lang;
using namespace cms;
using namespace std;


SimpleProducer::SimpleProducer( const std::string& brokerURI,
                                const std::string& destname, // The name of the queue or topic
                                bool useTopic,
                                bool sessionTransacted) {
   this->connection = NULL;
   this->session = NULL;
   this->destination = NULL;
   this->producer = NULL;
   this->useTopic = useTopic;
   this->sessionTransacted = sessionTransacted;
   this->brokerURI = brokerURI;
   this->destname = destname;
}

int SimpleProducer::Connect() {

   try {
      cleanup();

      // Create a ConnectionFactory
      auto_ptr<ConnectionFactory> connectionFactory(
         ConnectionFactory::createCMSConnectionFactory( brokerURI ) );


      // Create a Connection
      connection = connectionFactory->createConnection();


      connection->start();

      // Create a Session
      //      if( this->sessionTransacted ) {
      //         session = connection->createSession( Session::SESSION_TRANSACTED );
      //      } else {
               session = connection->createSession( Session::AUTO_ACKNOWLEDGE );
	       //      }

      // Create the destination (Topic or Queue)
      if( useTopic ) {
         destination = session->createTopic( destname );
      } else {
         destination = session->createQueue( destname );
      }

      // Create a MessageProducer from the Session to the Topic or Queue
      producer = session->createProducer( destination );
      producer->setDeliveryMode( DeliveryMode::NON_PERSISTENT );



   } catch ( CMSException& e ) {
     Error("SimpleProducer::Connect", "Could not connect " << e.getMessage() << endl << e.getStackTraceString() );
     return 1;
   }

   return 0;
}

// Produce some test messages
int SimpleProducer::Say(char *text, int count) {

      // Create the Thread Id String
      string threadIdStr = Integer::toString( Thread::getId() );

      if (!session) return 1;

      try {
         for( int ix = 0; ix < count; ++ix ){
            TextMessage* message = session->createTextMessage( text );
            
            if (!message) {
               Error("SimpleProducer::Say", "Could not create message " << ix);
               break;
            }

            message->setIntProperty( "Integer", ix );
            
            // Tell the producer to send the message
            if ((ix % 1000) == 0)
               Info(SimpleDebug::kLOW, "SimpleProducer::Say", "Sent message " << ix+1 << " from thread " << threadIdStr);

            producer->send( message );
            
            delete message;
         }
         
         
      }catch ( CMSException& e ) {
	Error("SimpleProducer::Say",
	      "CMSException catched while sending textmessage " << e.getStackTraceString());
      }

      return 0;

};



void SimpleProducer::cleanup(){
  char *fname = "SimpleProducer::cleanup";

   // Destroy resources.
   try{
      if( destination != NULL ) delete destination;
   }catch ( CMSException& e ) {
     Error(fname,
	   "CMSException catched while deleting destination " << e.getStackTraceString());
   }
   destination = NULL;

   try{
      if( producer != NULL ) delete producer;
   }catch ( CMSException& e )  {
     Error(fname,
	   "CMSException catched while deleting producer " << e.getStackTraceString());
   }
   producer = NULL;

   // Close open resources.
   try{
      if( session != NULL ) session->close();
      if( connection != NULL ) connection->close();
   }catch ( CMSException& e )  {
     Error(fname,
	   "CMSException catched while closing session/connection " << e.getStackTraceString());
   }

   try{
      if( session != NULL ) delete session;
   }catch ( CMSException& e )   {
     Error(fname,
	   "CMSException catched while deleting session " << e.getStackTraceString());
   }
   session = NULL;

   try{
      if( connection != NULL ) delete connection;
   }catch ( CMSException& e )  {
     Error(fname,
	   "CMSException catched while deleting connection " << e.getStackTraceString());
   }
   connection = NULL;
}

